# -*- coding: utf-8 -*-

import pprint
import cPickle
import numpy as np

from dataton_constants import *
import datetime
import re
from scipy import sparse
from sklearn import preprocessing

def build_X_y_from_custom_iter_ch_V3(
	custom_iter,
	n_customs,
	train_test = "train",
	add_bag_of_prods = True,
	add_bop_abs_dates = False,
	add_bop_rel_dates = False,
	add_last_prods = 0,
	add_lst_prd_abs_dates = False,
	add_lst_prd_rel_dates = False,
	scale_abs_dates_equal = False,
	scale_rel_dates_equal = False,
	orig_date = datetime.datetime(1950,1,1),
	abs_axis_norm = 0,
	rel_axis_norm = 1,
	rel_alpha = 24*3600,
	debug = False,
	):
	"""
	Created: 2-3-2017
	Modified: 14-3-2017
	Build the features matrix (X) and the labels vector (y) for an iterable
	of customers where each customer has a list of dictionaries with the 
	following entries:
	{
	"ID_Customer":str,
	"Socio_Demo_01":int,
	"Socio_Demo_02":int,
	"Socio_Demo_03":int,
	"Socio_Demo_04":int,
	"Socio_Demo_05":int,
	"Cod_Fecha":datetime,
	"Cod_Prod":int
	}
	Some clients are filtered out based on several criteria.

	custom_iter: iterable of customers.

	n_customs: number of customers.
	
	add_bag_of_prods: If True, add the Bag-of-products vector.

	add_bop_abs_dates: If True, add a vector of 94 features where the
	purchased products have a non-zero entry indicating the age of the 
	transaction.

	add_bop_rel_dates: If True, add a vector of 94 features where the 
	purchased products have a non-zero entry indicating the relative 
	time interval between transactions.

	add_last_prods: Number of the last products to add as features.

	add_lst_prd_abs_dates: If True, add a feature (values between 0 and 1) that takes
	into account the age of each transaction for the P-n products.
	
	add_lst_prd_rel_dates: If True, add a feature (values between 0 and 1) that takes
	into account the time interval from the last transaction. Applied to the 
	P-n products.
	"""
	## Compute the number of features and where in the features vector start
	## each binarized feature (that is stored in the offset variables).
	date_feats_abs = []
	date_feats_rel = []
	socdem_feats = 4*3+1+3
	n_features = socdem_feats
	if add_bag_of_prods:
		bop_offset = n_features
		n_features += 94
	if add_bop_abs_dates:
		abs_offset = n_features
		date_feats_abs.extend(range(abs_offset,abs_offset+94))
		n_features += 94
	if add_bop_rel_dates:
		rel_offset = n_features
		date_feats_rel.extend(range(rel_offset,rel_offset+94))
		n_features += 94
	lp_offset = n_features
	n_features += add_last_prods*94
	if add_lst_prd_abs_dates:
		lp_abs_offset = n_features
		date_feats_abs.extend(range(lp_abs_offset,lp_abs_offset+add_last_prods))
		n_features += add_last_prods
	if add_lst_prd_rel_dates:
		lp_rel_offset = n_features
		date_feats_rel.extend(range(lp_rel_offset,lp_rel_offset+add_last_prods-1))
		n_features += add_last_prods-1

	## Build sparse feature matrix
	X = sparse.lil_matrix((n_customs,n_features))
	y = []
	cust_dct = {}

	## If computing the features matrix of the train set, take every
	## transaction but the last one, which will be the label. 
	## If computing the features matrix of the test set take all the 
	## transactions.
	if train_test == "train":
		lim = -1
	elif train_test == "test":
		lim = None
	else:
		raise Exception("train_test choice not supported")

	## Start iterating through the customers and build the sparse feature
	## matrix by storing only the positions of the non-zero features and
	## their values in "pos" and "vals" vectors.
	for i, c in enumerate(custom_iter):
		## Several debug prints are scattered through this function
		if debug:
			pprint.pprint(c)
		if i % 10000 == 0:
			print "%d/%d customers processed"%(i,n_customs)
		## Sort the transactions first by date and then by Cod_Prod
		data = sorted(c["fields"],key=lambda x:(x["Cod_Fecha"],x["Cod_Prod"]))

		pos = []
		vals = []
		## Extract sociodemo features and binarize them
		soc_dem = [data[0]["Socio_Demo_0%d"%j] for j in xrange(1,6)]
		for j in xrange(3):
			if soc_dem[j] < 5:
				pos.append(soc_dem[j]+4*j-1)
				vals.append(1)
		if soc_dem[3] == 1:
			pos.append(12)
			vals.append(1)
		if soc_dem[4] < 3:
			pos.append(13+soc_dem[4])
			vals.append(1)
		if debug:
			print "socdem: ", pos, vals
		## Extract Bag-of-products features
		if add_bag_of_prods:
			pos_val = [(prod_pos[prd["Cod_Prod"]] + bop_offset,1) for prd in data[:lim]]
			try:
				pos0, val0 = zip(*pos_val)
				pos.extend(pos0)
				vals.extend(val0)
			except ValueError:
				## If a customer with just one product is encountered, pos_val 
				## would be empty. Catch that exception and go on.
				pass
			if debug:
				print "BoP: ", pos, vals
		## Extract all the absolute dates and store them in a vector of 94 elements
		if add_bop_abs_dates:
			pos.extend([prod_pos[prd["Cod_Prod"]] + abs_offset for prd in data[:lim]])
			dts = [(prd["Cod_Fecha"] - orig_date).total_seconds() for prd in data[:lim]]
			vals.extend(dts)
			if debug:
				print "Abs: ", pos, vals, "Dts: ", dts
		## Extract all the relative dates and store them in a vector of 94 elements
		if add_bop_rel_dates:
			pos.extend([prod_pos[prd["Cod_Prod"]] + rel_offset for prd in data[1:lim]])
			dts0 = [prd["Cod_Fecha"] for prd in data[:lim]]
			dts = [(dts0[j+1]-dts0[j]).total_seconds() + rel_alpha for j in xrange(len(dts0)-1)]
			vals.extend(dts)
			if debug:
				print "Rel: ", pos, vals, "Dts: ", dts
		## Extract last products features
		if add_last_prods:
			if train_test == "train":
				pos_val = [(prod_pos[prd["Cod_Prod"]]+94*j+lp_offset,1) for j,prd in enumerate(reversed(data[-(1+add_last_prods):lim]))]
			elif train_test == "test":
				pos_val = [(prod_pos[prd["Cod_Prod"]]+94*j+lp_offset,1) for j,prd in enumerate(reversed(data[-(add_last_prods):lim]))]
			try:
				pos0, val0 = zip(*pos_val)
				pos.extend(pos0)
				vals.extend(val0)
			except ValueError:
				## If a customer with just one product is encountered, pos_val 
				## would be empty. Catch that exception and go on.
				pass
			if debug:
				print "LP: ", pos, vals
		## Extract the absolute dates of the "add_last_prods" products
		if add_lst_prd_abs_dates:
			pos.extend(range(lp_abs_offset,lp_abs_offset+add_last_prods))
			dts = np.zeros(add_last_prods)
			if train_test == "train":
				dts0 = [(prd["Cod_Fecha"] - orig_date).total_seconds() for prd in reversed(data[-(1+add_last_prods):lim])]
			elif train_test == "test":
				dts0 = [(prd["Cod_Fecha"] - orig_date).total_seconds() for prd in reversed(data[-(add_last_prods):lim])]
			dts[:len(dts0)] = dts0
			vals.extend(dts)
			if debug:
				print "LPAbs: ", pos, vals, "Dts: ", dts
		## Extract the relative dates of the "add_last_prods" products
		if add_lst_prd_rel_dates:
			pos.extend(range(lp_rel_offset,lp_rel_offset+add_last_prods-1))
			dts = np.zeros(add_last_prods-1)
			if train_test == "train":
				dts0 = [prd["Cod_Fecha"] for prd in reversed(data[-(1+add_last_prods):lim])]
			elif train_test == "test":
				dts0 = [prd["Cod_Fecha"] for prd in reversed(data[-(add_last_prods):lim])]
			dts1 = [(dts0[j]-dts0[j+1]).total_seconds() + rel_alpha for j in xrange(len(dts0)-1)]
			dts[:len(dts1)] = dts1
			vals.extend(dts)
			if debug:
				print "LPRel: ", pos, vals, "Dts: ", dts

		## Build labels vector
		if train_test == "train":
			y.append(prod_pos[data[-1]["Cod_Prod"]])
		## Build sparse feature matrix
		X[i,pos] = vals
		## Build dictionary with positions of customers
		cust_dct[data[0]["ID_Customer"]] = i
	if debug:
		#### DEBUG
		print X.shape, date_feats_abs, date_feats_rel
		####
	X = X.tocsr()
	## Standarize the dates features
	if scale_abs_dates_equal:
		maxdate = X[:,date_feats_abs].max()
		X[:,date_feats_abs] = rescale(X[:,date_feats_abs],0,maxdate,0.0,1.0)
	else:
		if len(date_feats_abs) > 0:
			X[:,date_feats_abs] = preprocessing.maxabs_scale(X[:,date_feats_abs],axis=abs_axis_norm)
	if scale_rel_dates_equal:
		maxdate = X[:,date_feats_rel].max()
		X[:,date_feats_rel] = rescale(X[:,date_feats_rel],0,maxdate,0.0,1.0)
	else:
		if len(date_feats_rel) > 0:
			X[:,date_feats_rel] = preprocessing.maxabs_scale(X[:,date_feats_rel],axis=rel_axis_norm)
	if train_test == "train":
		return X, np.array(y), cust_dct
	elif train_test == "test":
		return X, cust_dct

def get_customers_lst(
	load_path
	):
	"""
	Builds and returns a list of dictionaries with the following structure:
	{'_id': *ID_Customer*,
	'fields': [
			{
			u'Cod_Fecha': *Cod_Fecha in datetime format*,
			u'Cod_Prod': *Cod_Prod*,
			u'ID_Customer': *ID_Customer*,
			u'Socio_Demo_01': *Socio_Demo_01*,
			u'Socio_Demo_02': *Socio_Demo_02*,
			u'Socio_Demo_03': *Socio_Demo_02*,
			u'Socio_Demo_04': *Socio_Demo_02*,
			u'Socio_Demo_05': *Socio_Demo_02*,
			},
			...
		]
	}
	Where key 'fileds' stores the list of transactions of the customer.
	load_path: Path to the csv file.
	"""
	custom_dict = {}
	with open(load_path,"r") as f:
		header = re.split(r"[\s|]+",f.next())[:-1]
		for j, l in enumerate(f):
			if j % 100000 == 0:
				print "%d transactions processed"%j
			vals = l.split("|")
			for i in [1,3,4,5,6,7]:
				vals[i] = int(vals[i])
			vals[2] = datetime.datetime.strptime(vals[2],"%Y-%m")
			reg = {header[i]:vals[i] for i in xrange(8)}
			try:
				custom_dict[reg["ID_Customer"]]["fields"].append(reg)
			except KeyError:
				custom_dict[reg["ID_Customer"]] = {
					"_id":reg["ID_Customer"],
					"fields":[reg]
				}
	return custom_dict.values()

# if __name__ == "__main__":
	

##############################################################################
##############################################################################
## Executed scripts
##############################################################################
##############################################################################

## 23-3-2017, pruebas para generar matrices features directamente con fichero
## csv. FUNCIONA OK.

# load_path = "../../../data/train2.txt"
# cstm_lst = get_customers_lst(
# load_path
# )
# print "Customer list generated"
# X, y, cust_dct = build_X_y_from_custom_iter_ch_V3(
# cstm_lst,
# len(cstm_lst),
# add_bag_of_prods = True,
# add_bop_abs_dates = False,
# add_bop_rel_dates = False,
# add_last_prods = 6,
# add_lst_prd_abs_dates = True,
# add_lst_prd_rel_dates = True,
# scale_abs_dates_equal = False,
# scale_rel_dates_equal = False,
# orig_date = datetime.datetime(1950,1,1),
# abs_axis_norm = 0,
# rel_axis_norm = 0,
# rel_alpha = 24*3600,
# debug = False,
# )
# print X.shape

# with open("../../resultados_parciales/test/P06_BoP1_Abs0_Rel0_LPAbs1_LPRel1_SclAbsEq0_SclRelEq0_AbsAxNorm0_RelAxNorm0/X.p","rb") as f:
# 	X2 = cPickle.load(f)
# with open("../../resultados_parciales/test/P06_BoP1_Abs0_Rel0_LPAbs1_LPRel1_SclAbsEq0_SclRelEq0_AbsAxNorm0_RelAxNorm0/custom_pos_dct.p","rb") as f:
# 	cust_dct2 = cPickle.load(f)

# print X2.shape

# X = X.tolil()
# X2 = X2.tolil()

# for i in np.random.randint(X.shape[0],size=100):
# 	row1 = X[i]
# 	cst = [k for k,v in cust_dct.iteritems() if v == i][0]
# 	i2 = cust_dct2[cst]
# 	row2 = X2[i2]
# 	print (row1 != row2).todense().any()

##############################################################################